# Copyright 2021-2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Client for connecting to the IQM quantum computer server interface.

The :class:`Circuit` class represents quantum circuits to be executed, consisting of a list of
native quantum operations, each represented by an instance of the :class:`Instruction` class.
Different Instruction types are distinguished by their :attr:`~Instruction.name`.
Each Instruction type acts on a number of :attr:`~Instruction.qubits`, and expects certain
:attr:`~Instruction.args`.


Instructions
============

We currently support three native instruction types:

================ =========== ====================================== ===========
name             # of qubits args                                   description
================ =========== ====================================== ===========
measurement      >= 1        ``key: str``                           Measurement in the Z basis.
phased_rx        1           ``angle_t: float``, ``phase_t: float`` Phased x-rotation gate.
cz               2                                                  Controlled-Z gate.
barrier          >= 2                                               Barrier instruction.
================ =========== ====================================== ===========

Measurement
-----------

Measurement in the computational (Z) basis. The measurement results are the output of the circuit.
Takes one string argument, ``key``, denoting the measurement key the results are labeled with.
All the measurement keys in a circuit must be unique.
Each qubit may only be measured once.
The measurement must be the last operation on each qubit, i.e. it cannot be followed by gates.

Example: ``Instruction(name='measurement', qubits=['alice', 'bob', 'charlie'], args={'key': 'm1'})``


Phased Rx
---------

Phased x-rotation gate, i.e. an x-rotation conjugated by a z-rotation.
Takes two arguments, the rotation angle ``angle_t`` and the phase angle ``phase_t``,
both measured in units of full turns (:math:`2\pi` radians).
The gate is represented in the standard computational basis by the matrix

.. math::
    R(\theta, \phi) = \exp(-i (X \cos (2 \pi \; \phi) + Y \sin (2 \pi \; \phi)) \: \pi \; \theta)
    = R_z(\phi) R_x(\theta) R_z^\dagger(\phi),

where :math:`\theta` = ``angle_t``, :math:`\phi` = ``phase_t``,
and :math:`X` and :math:`Y` are Pauli matrices.

Example: ``Instruction(name='phased_rx', qubits=['bob'], args={'angle_t': 0.7, 'phase_t': 0.25})``


CZ
--

Controlled-Z gate. Represented in the standard computational basis by the matrix

.. math:: \text{CZ} = \text{diag}(1, 1, 1, -1).

Symmetric wrt. the qubits it's acting on. Takes no arguments.

Example: ``Instruction(name='cz', qubits=['alice', 'bob'], args={})``


Barrier
-------

Barriers ensure that all operations after the barrier on the qubit subsystems spanned by
the barrier are only executed when all the operations before the barrier have been completed.

Example: ``Instruction(name='barrier', qubits=['alice', 'bob'], args={})``


Circuit output
==============

The :class:`RunResult` class represents the results of the quantum circuit execution.
If the run succeeded, :attr:`RunResult.measurements` contains the output of the batch of circuits,
consisting of the results of the measurement operations in each circuit.
It is a list of dictionaries, where each dict maps each measurement key to a 2D array of measurement
results, represented as a nested list.
``RunResult.measurements[circuit_index][key][shot][qubit_index]`` is the result of measuring the
``qubit_index``'th qubit in measurement operation ``key`` in the shot ``shot`` in the
``circuit_index``'th circuit of the batch.

The results are nonnegative integers representing the computational basis state (for qubits, 0 or 1)
that was the measurement outcome.

----
"""
from __future__ import annotations

import json
import os
import time
import warnings
from base64 import b64decode
from datetime import datetime
from enum import Enum
from posixpath import join
from typing import Any, Optional, Union
from uuid import UUID

import requests
from pydantic import BaseModel, Field

REQUESTS_TIMEOUT = 10

DEFAULT_TIMEOUT_SECONDS = 900
SECONDS_BETWEEN_CALLS = 1
REFRESH_MARGIN_SECONDS = 5

AUTH_CLIENT_ID = 'iqm_client'
AUTH_REALM = 'cortex'


class ClientConfigurationError(RuntimeError):
    """Wrong configuration provided.
    """


class ClientAuthenticationError(RuntimeError):
    """Something went wrong with user authentication.
    """


class CircuitExecutionError(RuntimeError):
    """Something went wrong on the server.
    """


class APITimeoutError(CircuitExecutionError):
    """Exception for when executing a task on the server takes too long.
    """


class Status(str, Enum):
    """
    Status of a task.
    """
    PENDING = 'pending'
    READY = 'ready'
    FAILED = 'failed'


class Instruction(BaseModel):
    """An instruction in a quantum circuit.
    """
    name: str = Field(..., description='name of the quantum operation', example='measurement')
    'name of the quantum operation'
    qubits: list[str] = Field(
        ...,
        description='names of the logical qubits the operation acts on',
        example=['alice'],
    )
    'names of the logical qubits the operation acts on'
    args: dict[str, Any] = Field(
        ...,
        description='arguments for the operation',
        example={'key': 'm'},
    )
    'arguments for the operation'


class Circuit(BaseModel):
    """Quantum circuit to be executed.
    """
    name: str = Field(..., description='name of the circuit', example='test circuit')
    'name of the circuit'
    instructions: list[Instruction] = Field(..., description='instructions comprising the circuit')
    'instructions comprising the circuit'

    def all_qubits(self) -> set[str]:
        """Return the names of all qubits in the circuit.
        """
        qubits = set()
        for instruction in self.instructions:
            qubits.update(instruction.qubits)
        return qubits


class SingleQubitMapping(BaseModel):
    """Mapping of a logical qubit name to a physical qubit name.
    """
    logical_name: str = Field(..., description='logical qubit name', example='alice')
    'logical qubit name'
    physical_name: str = Field(..., description='physical qubit name', example='QB1')
    'physical qubit name'


def serialize_qubit_mapping(qubit_mapping: dict[str, str]) -> list[SingleQubitMapping]:
    """Serializes a qubit mapping dict into the corresponding IQM data transfer format.

    Args:
        qubit_mapping: mapping from logical to physical qubit names

    Returns:
        data transfer object representing the mapping
    """
    return [SingleQubitMapping(logical_name=k, physical_name=v) for k, v in qubit_mapping.items()]


class RunRequest(BaseModel):
    """Request for an IQM quantum computer to execute a batch of quantum circuits.

    Note: all circuits in a batch must measure the same qubits otherwise batch execution fails.
    """
    circuits: list[Circuit] = Field(..., description='batch of quantum circuit(s) to execute')
    'batch of quantum circuit(s) to execute'
    settings: Optional[dict[str, Any]] = Field(
        None,
        description='IQM hardware settings and calibration data, or None to use the latest calibration set'
    )
    'IQM hardware settings and calibration data, or None to use the latest calibration set'
    calibration_set_id: Optional[int] = Field(
        None,
        description='ID of the calibration set to use, or None to use the latest calibration set'
    )
    'ID of the calibration set to use, or None to use the latest calibration set'
    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
        None,
        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
    )
    'mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
    shots: int = Field(..., description='how many times to execute each circuit in the batch')
    'how many times to execute each circuit in the batch'


CircuitMeasurementResults = dict[str, list[list[int]]]
"""Measurement results from a single circuit. For each measurement operation in the circuit,
maps the measurement key to the corresponding results. The outer list elements correspond to shots,
and the inner list elements to the qubits measured in the measurement operation."""


class Metadata(BaseModel):
    """Metadata belonging to a job sumission"""
    shots: int = Field(..., description='how many times to execute each circuit in the batch')
    'how many times to execute each circuit in the batch'
    qubit_mapping: Optional[list[SingleQubitMapping]] = Field(
        None,
        description='mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
    )
    'mapping of logical qubit names to physical qubit names, or None if using physical qubit names'
    circuits: list[Circuit] = Field(..., description='batch of quantum circuit(s) to execute')
    'batch of quantum circuit(s) to execute'
    calibration_set_id: Optional[int] = Field(
        None,
        description='ID of the calibration set used, or None if settings were specified'
    )


class RunResult(BaseModel):
    """Results of executing a batch of circuit(s).

    * ``measurements`` is present iff the status is ``'ready'``.
    * ``message`` carries additional information for the ``'failed'`` status.
    * If the status is ``'pending'``, ``measurements`` and ``message`` are ``None``.
    """
    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
    "current status of the run, in ``{'pending', 'ready', 'failed'}``"
    measurements: Optional[list[CircuitMeasurementResults]] = Field(
        None,
        description='if the run has finished successfully, the measurement results for the circuit(s)'
    )
    'if the run has finished successfully, the measurement results for the circuit(s)'
    message: Optional[str] = Field(None, description='if the run failed, an error message')
    'if the run failed, an error message'
    metadata: Metadata = Field(..., description='metadata about the underlying job request')
    'metadata about the underlying job request'
    warnings: Optional[list[str]] = Field(None, description='list of warning messages')
    'list of warning messages'

    @staticmethod
    def from_dict(inp: dict[str, Union[str, dict]]) -> RunResult:
        """Parses the result from a dict.

        Args:
            inp: value to parse, has to map to RunResult

        Returns:
            parsed run result

        """
        input_copy = inp.copy()
        return RunResult(status=Status(input_copy.pop('status')), **input_copy)


class RunStatus(BaseModel):
    """Status of a batchcircuit execution request."""
    status: Status = Field(..., description="current status of the run, in ``{'pending', 'ready', 'failed'}``")
    "current status of the run, in ``{'pending', 'ready', 'failed'}``"
    message: Optional[str] = Field(None, description='if the run failed, an error message')
    'if the run failed, an error message'
    warnings: Optional[list[str]] = Field(None, description='list of warning messages')
    'list of warning messages'

    @staticmethod
    def from_dict(inp: dict[str, Union[str, dict]]) -> RunStatus:
        """Parses the result from a dict.

        Args:
            inp: value to parse, has to map to RunResult

        Returns:
            parsed run result

        """
        input_copy = inp.copy()
        return RunStatus(status=Status(input_copy.pop('status')), **input_copy)


class GrantType(str, Enum):
    """
    Type of token request.
    """
    PASSWORD = 'password'
    REFRESH = 'refresh_token'


class AuthRequest(BaseModel):
    """Request sent to authentication server for access token and refresh token, or for terminating the session.

    * Token request with grant type ``'password'`` starts a new session in the authentication server.
      It uses fields ``client_id``, ``grant_type``, ``username`` and ``password``.
    * Token request with grant type ``'refresh_token'`` is used for maintaining an existing session.
      It uses field ``client_id``, ``grant_type``, ``refresh_token``.
    * Logout request uses only fields ``client_id`` and ``refresh_token``.

    """
    client_id: str = Field(..., description='name of the client for all request types')
    'name of the client for all request types'
    grant_type: Optional[GrantType] = Field(
        None,
        description="type of token request, in ``{'password', 'refresh_token'}``"
    )
    "type of token request, in ``{'password', 'refresh_token'}``"
    username: Optional[str] = Field(None, description="username for grant type ``'password'``")
    "username for grant type ``'password'``"
    password: Optional[str] = Field(None, description="password for grant type ``'password'``")
    "password for grant type ``'password'``"
    refresh_token: Optional[str] = Field(
        None,
        description="refresh token for grant type ``'refresh_token'`` and logout request")
    "refresh token for grant type ``'refresh_token'`` and logout request"


class Credentials(BaseModel):
    """Credentials and tokens for maintaining a session with the authentication server.

    * Fields ``auth_server_url``, ``username`` and ``password`` are provided by the user.
    * Fields ``access_token`` and ``refresh_token`` are loaded from the authentication server and
      refreshed periodically.
    """
    auth_server_url: str = Field(..., description='Base URL of the authentication server')
    'Base URL of the authentication server'
    username: str = Field(..., description='username for logging in to the server')
    'username for logging in to the server'
    password: str = Field(..., description='password for logging in to the server')
    'password for logging in to the server'
    access_token: Optional[str] = Field(None, description='current access token of the session')
    'current access token of the session'
    refresh_token: Optional[str] = Field(None, description='current refresh token of the session')
    'current refresh token of the session'


class ExternalToken(BaseModel):
    """Externally managed token for maintaining a session with the authentication server.

    * Fields ``auth_server_url`` and ``access_token`` are loaded from an
      external resource, e.g. file generated by Cortex CLI's token manager.
    """

    auth_server_url: str = Field(..., description='Base URL of the authentication server')
    'Base URL of the authentication server'
    access_token: str = Field(None, description='current access token of the session')
    'current access token of the session'


def _get_credentials(credentials: dict[str, str]) -> Optional[Credentials]:
    """Try to obtain credentials, first from arguments, then from environment variables.

    Args:
        credentials: dict of credentials provided as arguments

    Returns:
        Credentials with token fields cleared, or None if ``auth_server_url`` was not set.
    """
    auth_server_url = credentials.get('auth_server_url') or os.environ.get('IQM_AUTH_SERVER')
    username = credentials.get('username') or os.environ.get('IQM_AUTH_USERNAME')
    password = credentials.get('password') or os.environ.get('IQM_AUTH_PASSWORD')
    if not auth_server_url:
        return None
    if not username or not password:
        raise ClientConfigurationError('Auth server URL is set but no username or password')
    return Credentials(auth_server_url=auth_server_url, username=username, password=password)


def _get_external_token(tokens_file: Optional[str] = None) -> Optional[ExternalToken]:
    """Try to obtain external token from a file, first by path provided, then by path from
    environment variable.

    Args:
        tokens_file: path to a JSON file containing tokens

    Returns:
        ExternalToken with non-empty auth_server_url and access_token fields,
        or None if ``tokens_file`` was not provided.
    """

    filepath = tokens_file or os.environ.get('IQM_TOKENS_FILE')

    if not filepath:
        return None

    try:
        with open(filepath, 'r', encoding='utf-8') as file:
            raw_data = file.read()
    except FileNotFoundError as error:
        raise ClientConfigurationError(f'File not found: {filepath}') from error

    try:
        json_data = json.loads(raw_data)
    except json.decoder.JSONDecodeError as error:
        raise ClientConfigurationError(f'Decoding JSON has failed, {error}') from error

    auth_server_url = json_data['auth_server_url']
    access_token = json_data['access_token']

    return ExternalToken(auth_server_url=auth_server_url, access_token=access_token)


def _time_left_seconds(token: str) -> int:
    """Check how much time is left until the token expires.

    Returns:
        Time left on token in seconds.
    """
    _, body, _ = token.split('.', 2)
    # Add padding to adjust body length to a multiple of 4 chars as required by base64 decoding
    body += '=' * (-len(body) % 4)
    exp_time = int(json.loads(b64decode(body)).get('exp', '0'))
    return max(0, exp_time - int(time.time()))


class IQMClient:
    """Provides access to IQM quantum computers.

    Args:
        url: Endpoint for accessing the server. Has to start with http or https.
        tokens_file: Optional path to a tokens file used for authentication.
            This can also be set in the IQM_TOKENS_FILE environment variable.
            If tokens_file is set, auth_server_url, username and password
            must not be set.

    Keyword Args:
        auth_server_url: Optional base URL of the authentication server.
            This can also be set in the IQM_AUTH_SERVER environment variable.
            If unset, requests will be sent unauthenticated.
        username: Optional username to log in to authentication server.
            This can also be set in the IQM_AUTH_USERNAME environment variable.
            Username must be set if ``auth_server_url`` is set.
        password: Optional password to log in to authentication server.
            This can also be set in the IQM_AUTH_PASSWORD environment variable.
            Password must be set if ``auth_server_url`` is set.
    """

    def __init__(
            self,
            url: str,
            tokens_file: Optional[str] = None,
            **credentials  # contains auth_server_url, username, password
    ):
        if not url.startswith(('http:', 'https:')):
            raise ClientConfigurationError(f'The URL schema has to be http or https. Incorrect schema in URL: {url}')
        if tokens_file and credentials:
            raise ClientConfigurationError('Either external token or credentials must be provided. Both were provided.')
        self._base_url = url
        self._external_token = _get_external_token(tokens_file)
        if not self._external_token:
            self._credentials = _get_credentials(credentials)
            self._update_tokens()

    def __del__(self):
        try:
            # try our best to close the auth session, doesn't matter if it fails,
            # refresh token will be re-issued for the same credentials or eventually expire
            if not self._external_token:
                self.close_auth_session()
        except Exception:  # pylint: disable=broad-except
            pass

    # pylint: disable=too-many-locals
    def submit_circuits(
            self,
            circuits: list[Circuit],
            *,
            qubit_mapping: Optional[dict[str, str]] = None,
            settings: Optional[dict[str, Any]] = None,
            calibration_set_id: Optional[int] = None,
            shots: int = 1,
    ) -> UUID:
        """Submits a batch of quantum circuits for execution on a quantum computer.

        Args:
            circuits: list of circuit to be executed
            qubit_mapping: Mapping of human-readable (logical) qubit names in to physical qubit names.
                Can be set to ``None`` if all ``circuits`` already use physical qubit names.
                Note that the ``qubit_mapping`` is used for all ``circuits``.
            settings: settings for the quantum computer
            calibration_set_id: ID of the calibration set to use instead of ``settings``
            shots: number of times ``circuit`` is executed

        Returns:
            ID for the created task. This ID is needed to query the status and the execution results.
        """
        serialized_qubit_mapping: Optional[list[SingleQubitMapping]] = None
        if qubit_mapping is not None:
            # check if qubit mapping is injective
            target_qubits = set(qubit_mapping.values())
            if not len(target_qubits) == len(qubit_mapping):
                raise ValueError('Multiple logical qubits map to the same physical qubit.')

            # check if qubit mapping covers all qubits in the circuits
            for circuit in circuits:
                circuit_qubits = circuit.all_qubits()
                diff = circuit_qubits - set(qubit_mapping.keys())
                if diff:
                    raise ValueError(f'The qubits {diff} are not found in the provided qubit mapping.')

            # check that all the physical qubit names in qubit_mapping are defined in the settings
            if settings is not None:
                physical_qubits = set(settings['subtrees'])  # pylint: disable=unsubscriptable-object
                diff = target_qubits - physical_qubits
                if diff:
                    raise ValueError(f'The physical qubits {diff} in the qubit mapping are not defined in settings.')

            serialized_qubit_mapping = serialize_qubit_mapping(qubit_mapping)

        # ``bearer_token`` can be ``None`` if cocos we're connecting does not use authentication
        bearer_token = self._get_bearer_token()

        data = RunRequest(
            qubit_mapping=serialized_qubit_mapping,
            circuits=circuits,
            settings=settings,
            calibration_set_id=calibration_set_id,
            shots=shots
        )

        headers = {'Expect': '100-Continue'}
        if bearer_token:
            headers['Authorization'] = bearer_token

        result = requests.post(
            join(self._base_url, 'jobs'),
            json=data.dict(exclude_none=True),
            headers=headers,
            timeout=REQUESTS_TIMEOUT
        )

        if result.status_code == 401:
            raise ClientConfigurationError(f'Authentication failed: {result.text}')

        result.raise_for_status()
        return UUID(result.json()['id'])

    def get_run(self, job_id: UUID) -> RunResult:
        """Query the status and results of the running task.

        Args:
            job_id: id of the task

        Returns:
            result of the run (can be pending)

        Raises:
            HTTPException: http exceptions
            CircuitExecutionError: IQM server specific exceptions
        """
        bearer_token = self._get_bearer_token()
        result = requests.get(
            join(self._base_url, 'jobs/', str(job_id)),
            headers=None if not bearer_token else {'Authorization': bearer_token},
            timeout=REQUESTS_TIMEOUT
        )
        result.raise_for_status()
        run_result = RunResult.from_dict(result.json())
        if run_result.warnings:
            for warning in run_result.warnings:
                warnings.warn(warning)
        if run_result.status == Status.FAILED:
            raise CircuitExecutionError(run_result.message)
        return run_result

    def get_run_status(self, job_id: UUID) -> RunStatus:
        """Query the status of the running task.

        Args:
            job_id: id of the task

        Returns:
            status of the run

        Raises:
            HTTPException: http exceptions
            CircuitExecutionError: IQM server specific exceptions
        """
        bearer_token = self._get_bearer_token()
        result = requests.get(
            join(self._base_url, 'jobs/', str(job_id), 'status'),
            headers=None if not bearer_token else {'Authorization': bearer_token},
            timeout=REQUESTS_TIMEOUT
        )
        result.raise_for_status()
        run_result = RunStatus.from_dict(result.json())
        if run_result.warnings:
            for warning in run_result.warnings:
                warnings.warn(warning)
        return run_result

    def wait_for_results(self, job_id: UUID, timeout_secs: float = DEFAULT_TIMEOUT_SECONDS) -> RunResult:
        """Poll results until run is ready, failed, or timed out.

        Args:
            job_id: id of the task to wait
            timeout_secs: how long to wait for a response before raising an APITimeoutError

        Returns:
            run result

        Raises:
            APITimeoutError: time exceeded the set timeout
        """
        start_time = datetime.now()
        while (datetime.now() - start_time).total_seconds() < timeout_secs:
            results = self.get_run(job_id)
            if results.status != Status.PENDING:
                return results
            time.sleep(SECONDS_BETWEEN_CALLS)
        raise APITimeoutError(f"The task didn't finish in {timeout_secs} seconds.")

    def close_auth_session(self) -> bool:
        """Terminate session with authentication server if there was one created.

        Returns:
            True iff session was successfully closed

        Raises:
            ClientAuthenticationError: if logout failed
            ClientAuthenticationError: if asked to close externally managed authentication session
        """
        # auth session is managed externally, unable to close it here
        if self._external_token:
            raise ClientAuthenticationError('Unable to close externally managed auth session')

        # no auth, nothing to close
        if self._credentials is None:
            return False

        # auth session wasn't started, nothing to close
        if not self._credentials.refresh_token:
            return False

        url = f'{self._credentials.auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout'
        data = AuthRequest(client_id=AUTH_CLIENT_ID, refresh_token=self._credentials.refresh_token)
        result = requests.post(url, data=data.dict(exclude_none=True), timeout=REQUESTS_TIMEOUT)
        if result.status_code not in [200, 204]:
            raise ClientAuthenticationError(f'Logout failed, {result.text}')
        self._credentials.access_token = None
        self._credentials.refresh_token = None
        return True

    def _get_bearer_token(self, retries: int = 1) -> Optional[str]:
        """Make a bearer token for Authorization header. If access token is about to expire refresh it first.

        Args:
            retries: number of times to try updating the tokens

        Returns:
            Bearer token, i.e. string containing prefix 'Bearer ' and the access token, or None if access token
            is not available.
        """
        if self._external_token:
            return f'Bearer {self._external_token.access_token}'
        if self._credentials is None or not self._credentials.access_token:
            return None
        if _time_left_seconds(self._credentials.access_token) > REFRESH_MARGIN_SECONDS:
            # Access token is still valid, no need to refresh
            return f'Bearer {self._credentials.access_token}'
        if retries < 1:
            return None
        self._update_tokens()
        return self._get_bearer_token(retries - 1)

    def _update_tokens(self):
        """Update access token and refresh token.

        Uses refresh token to request new tokens from authentication server.  If the refresh token has expired or
        is about to expire, starts a new session by requesting the tokens using username and password.

        Updated tokens are stored in ``self._credentials``.

        Raises:
            ClientAuthenticationError: updating the tokens failed
        """
        if self._credentials is None:
            return

        refresh_token = self._credentials.refresh_token
        if refresh_token and _time_left_seconds(refresh_token) > REFRESH_MARGIN_SECONDS:
            # Update tokens using existing refresh_token
            data = AuthRequest(
                client_id=AUTH_CLIENT_ID,
                grant_type=GrantType.REFRESH,
                refresh_token=refresh_token
            )
        else:
            # Update tokens using username and password
            data = AuthRequest(
                client_id=AUTH_CLIENT_ID,
                grant_type=GrantType.PASSWORD,
                username=self._credentials.username,
                password=self._credentials.password
            )

        url = f'{self._credentials.auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/token'
        result = requests.post(url, data=data.dict(exclude_none=True), timeout=REQUESTS_TIMEOUT)
        if result.status_code != 200:
            raise ClientAuthenticationError(f'Failed to update tokens, {result.text}')
        tokens = result.json()
        self._credentials.access_token = tokens.get('access_token')
        self._credentials.refresh_token = tokens.get('refresh_token')



# Copyright 2021-2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the IQM client.
"""
# pylint: disable=unused-argument
import pytest
import requests
from mockito import when
from requests import HTTPError

from iqm_client import (Circuit, ClientConfigurationError, IQMClient,
                        SingleQubitMapping, Status, serialize_qubit_mapping)
from tests.conftest import MockJsonResponse, existing_run, missing_run

REQUESTS_TIMEOUT = 10


def test_serialize_qubit_mapping():
    qubit_mapping = {'Alice': 'QB1', 'Bob': 'qubit_3', 'Charlie': 'physical 0'}
    assert serialize_qubit_mapping(qubit_mapping) == [
        SingleQubitMapping(logical_name='Alice', physical_name='QB1'),
        SingleQubitMapping(logical_name='Bob', physical_name='qubit_3'),
        SingleQubitMapping(logical_name='Charlie', physical_name='physical 0'),
    ]


def test_submit_circuits_returns_id(mock_server, settings_dict, base_url, sample_circuit):
    """
    Tests sending a circuit
    """
    client = IQMClient(base_url)
    job_id = client.submit_circuits(
        circuits=[Circuit.parse_obj(sample_circuit)],
        qubit_mapping={'Qubit A': 'QB1', 'Qubit B': 'QB2'},
        settings=settings_dict,
        shots=1000)
    assert job_id == existing_run


def test_submit_circuit_with_non_injective_qubit_mapping(mock_server, settings_dict, base_url, sample_circuit):
    """
    Test non-injective qubit mapping.
    """
    client = IQMClient(base_url)
    with pytest.raises(ValueError, match='Multiple logical qubits map to the same physical qubit'):
        client.submit_circuits(
            circuits=[Circuit.parse_obj(sample_circuit)],
            qubit_mapping={'Qubit A': 'QB1', 'Qubit B': 'QB1'},
            settings=settings_dict
        )


def test_submit_circuit_with_non_existing_qubits(mock_server, settings_dict, base_url, sample_circuit):
    """
    Test qubit mapping containing a physical qubit name not present in the settings.
    """
    client = IQMClient(base_url)
    with pytest.raises(ValueError, match="{'QB200'} in the qubit mapping are not defined in settings."):
        client.submit_circuits(
            circuits=[Circuit.parse_obj(sample_circuit)],
            qubit_mapping={'Qubit A': 'QB1', 'Qubit B': 'QB200'},
            settings=settings_dict)


def test_submit_circuit_with_incomplete_qubit_mapping(mock_server, settings_dict, base_url, sample_circuit):
    """
    Test the scenario when circuits contain qubit names that are not present in the provided qubit mapping.
    """
    client = IQMClient(base_url)
    with pytest.raises(ValueError, match="The qubits {'Qubit B'} are not found in the provided qubit mapping."):
        client.submit_circuits(
            circuits=[Circuit.parse_obj(sample_circuit)],
            qubit_mapping={'Qubit A': 'QB1'},
            settings=settings_dict)


def test_submit_circuits_without_settings_returns_id(mock_server, base_url, sample_circuit):
    """
    Tests sending a circuit without settings
    """
    client = IQMClient(base_url)
    job_id = client.submit_circuits(
        qubit_mapping={'Qubit A': 'QB1', 'Qubit B': 'QB2'},
        circuits=[Circuit.parse_obj(sample_circuit)],
        shots=1000)
    assert job_id == existing_run


def test_submit_circuits_with_calibration_set_id_returns_id(mock_server, base_url, calibration_set_id, sample_circuit):
    """
    Tests sending a circuit with calibration set id
    """
    client = IQMClient(base_url)
    job_id = client.submit_circuits(
        qubit_mapping={'Qubit A': 'QB1', 'Qubit B': 'QB2'},
        circuits=[Circuit.parse_obj(sample_circuit)],
        calibration_set_id=calibration_set_id,
        shots=1000)
    assert job_id == existing_run


def test_submit_circuits_without_qubit_mapping_returns_id(mock_server, settings_dict, base_url, sample_circuit):
    """
    Tests sending a circuit without qubit mapping
    """
    client = IQMClient(base_url)
    job_id = client.submit_circuits(
        circuits=[Circuit.parse_obj(sample_circuit)],
        settings=settings_dict,
        shots=1000)
    assert job_id == existing_run


def test_get_run_status_and_results_for_existing_run(mock_server, base_url, settings_dict, calibration_set_id):
    """
    Tests getting the run status
    """
    client = IQMClient(base_url)
    assert client.get_run(existing_run).status == Status.PENDING
    ready_run = client.get_run(existing_run)
    assert ready_run.status == Status.READY
    assert ready_run.measurements is not None
    assert ready_run.metadata.calibration_set_id == calibration_set_id


def test_get_run_status_for_existing_run(mock_server, base_url, settings_dict):
    """
    Tests getting the run status
    """
    client = IQMClient(base_url)
    assert client.get_run_status(existing_run).status == Status.PENDING
    ready_run = client.get_run_status(existing_run)
    assert ready_run.status == Status.READY


def test_get_run_status_and_results_for_missing_run(mock_server, base_url, settings_dict):
    """
    Tests getting a task that was not created
    """
    client = IQMClient(base_url)
    with pytest.raises(HTTPError):
        assert client.get_run(missing_run)


def test_get_run_status_for_missing_run(mock_server, base_url, settings_dict):
    """
    Tests getting a task that was not created
    """
    client = IQMClient(base_url)
    with pytest.raises(HTTPError):
        assert client.get_run_status(missing_run)


def test_waiting_for_results(mock_server, base_url, settings_dict):
    """
    Tests waiting for results for an existing task
    """
    client = IQMClient(base_url)
    assert client.wait_for_results(existing_run).status == Status.READY


def test_user_warning_is_emitted_when_warnings_in_response(base_url, settings_dict, capsys):
    client = IQMClient(base_url)
    msg = 'This is a warning msg'
    with when(requests).get(f'{base_url}/jobs/{existing_run}', headers=None, timeout=REQUESTS_TIMEOUT).thenReturn(
            MockJsonResponse(200, {'status': 'ready', 'warnings': [msg], 'metadata': {'shots': 42, 'circuits': []}})
    ):
        with pytest.warns(UserWarning, match=msg):
            client.get_run(existing_run)


def test_base_url_is_invalid(settings_dict):
    invalid_base_url = 'https//example.com'
    with pytest.raises(ClientConfigurationError) as exc:
        IQMClient(invalid_base_url)
    assert f'The URL schema has to be http or https. Incorrect schema in URL: {invalid_base_url}' == str(exc.value)


def test_tokens_file_not_found(settings_dict):
    base_url = 'https://example.com'
    tokens_file = '/home/iqm/tokens.json'
    with pytest.raises(ClientConfigurationError) as exc:
        IQMClient(base_url, tokens_file=tokens_file)
    assert f'File not found: {tokens_file}' == str(exc.value)


def test_tokens_and_credentials_combo_invalid(settings_dict, credentials):
    tokens_file = '/home/iqm/tokens.json'
    base_url = 'https://example.com'
    with pytest.raises(ClientConfigurationError) as exc:
        IQMClient(base_url, tokens_file=tokens_file, **credentials)
    assert 'Either external token or credentials must be provided. Both were provided.' == str(exc.value)



# Copyright 2021-2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mocks server calls for testing
"""

import json
import os
import time
from base64 import b64encode
from typing import Optional
from uuid import UUID, uuid4

import pytest
import requests
from mockito import expect, mock, unstub, when
from requests import HTTPError, Response

from iqm_client import AUTH_CLIENT_ID, AUTH_REALM, AuthRequest, GrantType

REQUESTS_TIMEOUT = 10

calibration_set_id_value = 24
existing_run = UUID('3c3fcda3-e860-46bf-92a4-bcc59fa76ce9')
missing_run = UUID('059e4186-50a3-4e6c-ba1f-37fe6afbdfc2')


@pytest.fixture()
def base_url():
    return 'https://example.com'


@pytest.fixture()
def credentials():
    return {
        'auth_server_url': 'some_auth_server',
        'username': 'some_username',
        'password': 'some_password',
    }


@pytest.fixture(scope='function')
def mock_server(base_url, sample_circuit):
    """
    Runs mocking separately for each test
    """
    generate_server_stubs(base_url, sample_circuit)
    yield  # running test function
    unstub()


@pytest.fixture
def settings_dict():
    """
    Reads and parses settings file into a dictionary
    """
    settings_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/settings.json'
    with open(settings_path, 'r', encoding='utf-8') as f:
        return json.loads(f.read())


@pytest.fixture()
def calibration_set_id():
    return calibration_set_id_value


@pytest.fixture
def tokens_dict():
    """
    Reads and parses tokens file into a dictionary
    """
    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
    with open(tokens_path, 'r', encoding='utf-8') as f:
        return json.loads(f.read())


@pytest.fixture
def sample_circuit():
    """
    A sample circuit for testing submit_circuit
    """
    return {
        'name': 'The circuit',
        'instructions': [
            {
                'name': 'cz',
                'qubits': [
                    'Qubit A',
                    'Qubit B'
                ],
                'args': {}
            },
            {
                'name': 'phased_rx',
                'qubits': [
                    'Qubit A'
                ],
                'args': {
                    'phase_t': 0.7,
                    'angle_t': 0.25
                }
            },
            {
                'name': 'measurement',
                'qubits': [
                    'Qubit A'
                ],
                'args': {
                    'output_label': 'A'
                }
            }
        ]
    }


class MockJsonResponse:
    def __init__(self, status_code: int, json_data: dict):
        self.status_code = status_code
        self.json_data = json_data

    @property
    def text(self):
        return json.dumps(self.json_data)

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError('')


def generate_server_stubs(base_url, sample_circuit):
    """
    Mocking some calls to the server by mocking 'requests'
    """
    when(requests).post(f'{base_url}/jobs', ...).thenReturn(
        MockJsonResponse(201, {'id': str(existing_run)})
    )

    when(requests).get(f'{base_url}/jobs/{existing_run}', ...).thenReturn(
        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': [sample_circuit]}})
    ).thenReturn(
        MockJsonResponse(
            200,
            {
                'status': 'ready',
                'measurements': [{'result': [[1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 1, 1], [1, 0, 1, 1]]}],
                'metadata': {'shots': 42, 'circuits': [sample_circuit], 'calibration_set_id': calibration_set_id_value}
            }
        )
    )

    when(requests).get(f'{base_url}/jobs/{existing_run}/status', ...).thenReturn(
        MockJsonResponse(200, {'status': 'pending'})
    ).thenReturn(
        MockJsonResponse(200, {'status': 'ready'})
    )

    # 'run was not created' response
    no_run_response = Response()
    no_run_response.status_code = 404
    no_run_response.reason = 'Run not found'
    when(requests).get(f'{base_url}/jobs/{missing_run}', ...).thenReturn(no_run_response)
    when(requests).get(f'{base_url}/jobs/{missing_run}/status', ...).thenReturn(no_run_response)


def prepare_tokens(
        access_token_lifetime: int,
        refresh_token_lifetime: int,
        previous_refresh_token: Optional[str] = None,
        status_code: int = 200,
        **credentials
) -> dict[str, str]:
    """Prepare tokens and set them to be returned for a token request.

    Args:
        access_token_lifetime: seconds from current time to access token expire time
        refresh_token_lifetime: seconds from current time to refresh token expire time
        previous_refresh_token: refresh token to be used in refresh request
        status_code: status code to return for token request
        credentials: dict containing auth_server_url, username and password

    Returns:
         Prepared tokens as a dict.
    """
    if previous_refresh_token is None:
        request_data = AuthRequest(
            client_id=AUTH_CLIENT_ID,
            grant_type=GrantType.PASSWORD,
            username=credentials['username'],
            password=credentials['password']
        )
    else:
        request_data = AuthRequest(
            client_id=AUTH_CLIENT_ID,
            grant_type=GrantType.REFRESH,
            refresh_token=previous_refresh_token
        )

    tokens = {
        'access_token': make_token('Bearer', access_token_lifetime),
        'refresh_token': make_token('Refresh', refresh_token_lifetime)
    }
    when(requests).post(
        f'{credentials["auth_server_url"]}/realms/{AUTH_REALM}/protocol/openid-connect/token',
        data=request_data.dict(exclude_none=True),
        timeout=REQUESTS_TIMEOUT
    ).thenReturn(MockJsonResponse(status_code, tokens))

    return tokens


def make_token(token_type: str, lifetime: int) -> str:
    """Encode given token type and expire time as a token.

    Args:
        token_type: 'Bearer' for access tokens, 'Refresh' for refresh tokens
        lifetime: seconds from current time to token's expire time

    Returns:
        Encoded token
    """
    empty = b64encode('{}'.encode('utf-8')).decode('utf-8')
    body = f'{{ "typ": "{token_type}", "exp": {int(time.time()) + lifetime} }}'
    body = b64encode(body.encode('utf-8')).decode('utf-8')
    return f'{empty}.{body}.{empty}'


def expect_status_request(url: str, access_token: Optional[str], times: int = 1) -> UUID:
    """Prepare for status request.

    Args:
        url: server URL for the status request
        access_token: access token to use in Authorization header
            If not set, expect request to have no Authorization header
        times: number of times the status request is expected to be made

    Returns:
        Expected job ID to be used in the request
    """
    job_id = uuid4()
    headers = None if access_token is None else {'Authorization': f'Bearer {access_token}'}
    expect(requests, times=times).get(f'{url}/jobs/{job_id}', headers=headers, timeout=REQUESTS_TIMEOUT).thenReturn(
        MockJsonResponse(200, {'status': 'pending', 'metadata': {'shots': 42, 'circuits': []}})
    )
    return job_id


def expect_logout(auth_server_url: str, refresh_token: str):
    """Prepare for logout request.

    Args:
        auth_server_url: base URL of the authentication server
        refresh_token: refresh token expected to be used in the request
    """
    request_data = AuthRequest(client_id=AUTH_CLIENT_ID, refresh_token=refresh_token)
    expect(requests, times=1).post(
        f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout',
        data=request_data.dict(exclude_none=True),
        timeout=REQUESTS_TIMEOUT
    ).thenReturn(
        mock({'status_code': 204, 'text': '{}'})
    )



# Copyright 2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Script to publish JSON schemas for the Pydantic models used by the HTTP API.
"""
import json
import os
from typing import Any

from iqm_client import BaseModel, RunRequest

try:
    from iqm_client import __version__ as version
except ImportError:
    pass

# schemas to generate
SCHEMAS = {
    'run_request_schema': RunRequest,
}


def generate_json_schema(cls: type[BaseModel], filename: str) -> dict[str, Any]:
    """Generate a JSON schema dictionary from the given Pydantic model.

    Args:
        cls: pydantic model for which the JSON schema should be generated.
        filename: filename of the JSON schema.
    """
    json_schema = cls.schema()

    return {
        # JSON Schema version used
        '$schema': 'https://json-schema.org/draft/2020-12/schema',
        # URI containing the version of the generated schema
        '$id': f'https://iqm-finland.github.io/iqm-client/{filename}',
        **json_schema,
    }


def save_json_schemas_to_docs() -> None:
    """Save the JSON schemas in the docs folder.
    """
    for filename_prefix, cls in SCHEMAS.items():
        filename = f"{filename_prefix}_v{version}.json"
        json_schema_path = os.path.join(os.getcwd(), 'docs', filename)
        with open(json_schema_path, 'w', encoding='utf-8') as f:
            f.write(json.dumps(generate_json_schema(cls, filename), indent=2))


if __name__ == '__main__':
    save_json_schemas_to_docs()



# Copyright 2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for user authentication and token management in IQM client.
"""

import os

import pytest
from mockito import unstub

from iqm_client import ClientAuthenticationError, Credentials, IQMClient
from tests.conftest import expect_logout, expect_status_request, prepare_tokens


def test_get_initial_tokens_with_credentials_from_arguments(base_url, credentials):
    """
    Tests that if the client is initialized with credentials, they are used correctly
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    expected_credentials = Credentials(
        access_token=tokens['access_token'],
        refresh_token=tokens['refresh_token'],
        **credentials
    )
    client = IQMClient(base_url, **credentials)
    assert client._credentials == expected_credentials
    unstub()


def test_get_initial_tokens_with_credentials_from_env_variables(base_url, credentials, monkeypatch):
    """
    Tests that credentials are read from environment variables if they are not given as arguments
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    expected_credentials = Credentials(
        access_token=tokens['access_token'],
        refresh_token=tokens['refresh_token'],
        **credentials
    )
    monkeypatch.setenv('IQM_AUTH_SERVER', credentials['auth_server_url'])
    monkeypatch.setenv('IQM_AUTH_USERNAME', credentials['username'])
    monkeypatch.setenv('IQM_AUTH_PASSWORD', credentials['password'])
    client = IQMClient(base_url)
    assert client._credentials == expected_credentials
    unstub()


def test_add_authorization_header_when_credentials_are_provided(base_url, credentials):
    """
    Tests that requests are sent with Authorization header when credentials are provided
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    job_id = expect_status_request(base_url, tokens['access_token'])
    client = IQMClient(base_url, **credentials)
    result = client.get_run(job_id)
    assert result.status == 'pending'
    unstub()


def test_add_authorization_header_when_external_token_is_provided(base_url, tokens_dict):
    """
    Tests that requests are sent with Authorization header when credentials are provided
    """
    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
    job_id = expect_status_request(base_url, tokens_dict['access_token'])
    client = IQMClient(base_url, tokens_file=tokens_path)
    result = client.get_run(job_id)
    assert result.status == 'pending'
    unstub()


def test_no_authorization_header_when_credentials_are_not_provided(base_url):
    """
    Tests that requests are sent without Authorization header when no credentials are provided
    """
    job_id = expect_status_request(base_url, None)
    client = IQMClient(base_url)
    result = client.get_run(job_id)
    assert result.status == 'pending'
    unstub()


def test_raises_client_authentication_error_if_authentication_fails(base_url, credentials):
    """
    Tests that authentication failure raises ClientAuthenticationError
    """
    prepare_tokens(300, 3600, status_code=401, **credentials)
    with pytest.raises(ClientAuthenticationError):
        IQMClient(base_url, **credentials)
    unstub()


def test_access_token_is_not_refreshed_if_it_has_not_expired(base_url, credentials):
    """
    Test that access token is not refreshed if it has not expired
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    client = IQMClient(base_url, **credentials)
    assert client._credentials.access_token == tokens['access_token']

    job_id = expect_status_request(base_url, tokens['access_token'], 3)
    client.get_run(job_id)
    client.get_run(job_id)
    client.get_run(job_id)


def test_expired_access_token_is_refreshed_automatically(base_url, credentials):
    """
    Test that access token is refreshed automatically if it has expired
    """
    initial_tokens = prepare_tokens(-300, 3600, **credentials)  # expired initial access token
    refreshed_tokens = prepare_tokens(300, 4200, initial_tokens['refresh_token'], **credentials)
    job_id = expect_status_request(base_url, refreshed_tokens['access_token'])

    # Check initial access token
    client = IQMClient(base_url, **credentials)
    assert client._credentials.access_token == initial_tokens['access_token']

    # Check that assert token is refreshed
    result = client.get_run(job_id)
    assert client._credentials.access_token == refreshed_tokens['access_token']
    assert result.status == 'pending'

    unstub()


def test_start_new_session_when_refresh_token_has_expired(base_url, credentials):
    """
    Test that a new session is started automatically if refresh token has expired
    """
    initial_tokens = prepare_tokens(-3600, -300, **credentials)  # expired initial access token and refresh token

    client = IQMClient(base_url, **credentials)
    assert client._credentials.access_token == initial_tokens['access_token']
    assert client._credentials.refresh_token == initial_tokens['refresh_token']

    refreshed_tokens = prepare_tokens(300, 3600, **credentials)  # refreshed access token and refresh token
    job_id = expect_status_request(base_url, refreshed_tokens['access_token'])
    result = client.get_run(job_id)
    assert client._credentials.access_token == refreshed_tokens['access_token']
    assert result.status == 'pending'

    unstub()


def test_tokens_are_cleared_at_logout(base_url, credentials):
    """
    Tests that calling ``close`` will terminate the session and clear tokens
    """
    initial_tokens = prepare_tokens(300, 3600, **credentials)
    expect_logout(credentials['auth_server_url'], initial_tokens['refresh_token'])

    client = IQMClient(base_url, **credentials)
    assert client._credentials.access_token == initial_tokens['access_token']
    assert client._credentials.refresh_token == initial_tokens['refresh_token']

    client.close_auth_session()
    assert client._credentials.access_token is None
    assert client._credentials.refresh_token is None

    unstub()


def test_cannot_close_external_auth_session(base_url):
    """
    Tests that calling ``close_auth_session`` while initialized with an external auth session
    raises ClientAuthenticationError
    """
    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
    client = IQMClient(base_url, tokens_file=tokens_path)
    with pytest.raises(ClientAuthenticationError) as exc:
        client.close_auth_session()
    assert 'Unable to close externally managed auth session' == str(exc.value)


def test_logout_on_client_destruction(base_url, credentials):
    """
    Tests that client is trying to terminate the authentication session on destruction
    """
    initial_tokens = prepare_tokens(300, 3600, **credentials)
    expect_logout(credentials['auth_server_url'], initial_tokens['refresh_token'])

    client = IQMClient(base_url, **credentials)
    assert client._credentials.access_token == initial_tokens['access_token']
    assert client._credentials.refresh_token == initial_tokens['refresh_token']

    del client

    unstub()



# Copyright 2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests to check if the generated json schema validates run requests.
"""
# pylint: disable=redefined-outer-name
from copy import deepcopy

import pytest
from jsonschema import ValidationError, validate

from docs.generate_json_schemas import generate_json_schema
from iqm_client import RunRequest


@pytest.fixture
def sample_valid_run_request(sample_circuit):
    """
    Returns valid example run request.
    """
    return RunRequest(
        circuits=[sample_circuit],
        settings={},
        calibration_set_id=24,
        shots=1000,
        qubit_mapping=[{'logical_name': 'q1', 'physical_name': 'qubit_1'}]
    ).dict()


@pytest.fixture
def sample_invalid_run_request(sample_valid_run_request):
    """
    Returns invalid example run request.
    """
    invalid_run_request = deepcopy(sample_valid_run_request)
    invalid_run_request['shots'] = 'not_a_number'
    return invalid_run_request


def test_jsonschema_validates_run_requests(sample_valid_run_request):
    """
    Tests that the generated json schema validates valid run requests.
    """
    json_schema = generate_json_schema(RunRequest, '')
    validate(schema=json_schema, instance=sample_valid_run_request)


def test_jsonschema_throws_validation_errors(sample_invalid_run_request):
    """
    Tests that the generated json schema rejects invalid run requests.
    """
    json_schema = generate_json_schema(RunRequest, '')
    with pytest.raises(ValidationError):
        validate(schema=json_schema, instance=sample_invalid_run_request)
